/*
 * i386 register context support
 *
 * Copyright (C) 1999 Alexandre Julliard
 */

#include "config.h"

#ifdef __i386__

#include <assert.h>
#include <errno.h>
#ifdef HAVE_SYS_REG_H
#include <sys/reg.h>
#endif
#include <unistd.h>
#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif
#ifdef HAVE_SYS_PTRACE_H
# include <sys/ptrace.h>
#endif
#ifdef HAVE_SYS_PARAM_H
# include <sys/param.h>
#endif

#include "winbase.h"
#include "thread.h"
#include "process.h"
#include "request.h"

#ifndef PTRACE_PEEKUSER
#  ifdef PTRACE_PEEKUSR /* libc5 uses USR not USER */
#    define PTRACE_PEEKUSER PTRACE_PEEKUSR
#  else
#    define PTRACE_PEEKUSER PT_READ_U
#  endif
#endif

#ifndef PTRACE_POKEUSER
#  ifdef PTRACE_POKEUSR /* libc5 uses USR not USER */
#    define PTRACE_POKEUSER PTRACE_POKEUSR
#  else
#    define PTRACE_POKEUSER PT_WRITE_U
#  endif
#endif

#ifndef PTRACE_GETREGS
#define PTRACE_GETREGS PT_GETREGS
#endif
#ifndef PTRACE_GETFPREGS
#define PTRACE_GETFPREGS PT_GETFPREGS
#endif
#ifndef PTRACE_SETREGS
#define PTRACE_SETREGS PT_SETREGS
#endif
#ifndef PTRACE_SETFPREGS
#define PTRACE_SETFPREGS PT_SETFPREGS
#endif

#ifdef linux
#ifdef HAVE_SYS_USER_H
# include <sys/user.h>
#endif

/* user_regs definitions from asm/user.h */
struct kernel_user_regs_struct
{
    long ebx, ecx, edx, esi, edi, ebp, eax;
    unsigned short ds, __ds, es, __es;
    unsigned short fs, __fs, gs, __gs;
    long orig_eax, eip;
    unsigned short cs, __cs;
    long eflags, esp;
    unsigned short ss, __ss;
};

/* debug register offset in struct user */
#define DR_OFFSET(dr) ((int)((((struct user *)0)->u_debugreg) + (dr)))

/* retrieve a debug register */
static inline int get_debug_reg( int tid, int num, DWORD *data )
{
    int res = ptrace( PTRACE_PEEKUSER, tid, DR_OFFSET(num), 0 );
    if ((res == -1) && errno)
    {
        file_set_error();
        return -1;
    }
    *data = res;
    return 0;
}

/* retrieve a thread context */
void get_thread_context( struct thread *thread, unsigned int flags, CONTEXT *context )
{
    int tid = thread->unix_tid_or_pid;
    if (flags & CONTEXT_FULL)
    {
        struct kernel_user_regs_struct regs;
        if (ptrace( PTRACE_GETREGS, tid, 0, &regs ) == -1) goto error;
        if (flags & CONTEXT_INTEGER)
        {
            context->Eax = regs.eax;
            context->Ebx = regs.ebx;
            context->Ecx = regs.ecx;
            context->Edx = regs.edx;
            context->Esi = regs.esi;
            context->Edi = regs.edi;
        }
        if (flags & CONTEXT_CONTROL)
        {
            context->Ebp    = regs.ebp;
            context->Esp    = regs.esp;
            context->Eip    = regs.eip;
            context->SegCs  = regs.cs;
            context->SegSs  = regs.ss;
            context->EFlags = regs.eflags;
        }
        if (flags & CONTEXT_SEGMENTS)
        {
            context->SegDs = regs.ds;
            context->SegEs = regs.es;
            context->SegFs = regs.fs;
            context->SegGs = regs.gs;
        }
    }
    if (flags & CONTEXT_DEBUG_REGISTERS)
    {
        if (get_debug_reg( tid, 0, &context->Dr0 ) == -1) goto error;
        if (get_debug_reg( tid, 1, &context->Dr1 ) == -1) goto error;
        if (get_debug_reg( tid, 2, &context->Dr2 ) == -1) goto error;
        if (get_debug_reg( tid, 3, &context->Dr3 ) == -1) goto error;
        if (get_debug_reg( tid, 6, &context->Dr6 ) == -1) goto error;
        if (get_debug_reg( tid, 7, &context->Dr7 ) == -1) goto error;
    }
    if (flags & CONTEXT_FLOATING_POINT)
    {
        /* we can use context->FloatSave directly as it is using the */
        /* correct structure (the same as fsave/frstor) */
        if (ptrace( PTRACE_GETFPREGS, tid, 0, &context->FloatSave ) == -1) goto error;
        context->FloatSave.Cr0NpxState = 0;  /* FIXME */
    }
    return;
 error:
    file_set_error();
}


/* set a thread context */
void set_thread_context( struct thread *thread, unsigned int flags, const CONTEXT *context )
{
    int tid = thread->unix_tid_or_pid;
    if (flags & CONTEXT_FULL)
    {
        struct kernel_user_regs_struct regs;

        /* need to preserve some registers (at a minimum orig_eax must always be preserved) */
        if (ptrace( PTRACE_GETREGS, tid, 0, &regs ) == -1) goto error;

        if (flags & CONTEXT_INTEGER)
        {
            regs.eax = context->Eax;
            regs.ebx = context->Ebx;
            regs.ecx = context->Ecx;
            regs.edx = context->Edx;
            regs.esi = context->Esi;
            regs.edi = context->Edi;
        }
        if (flags & CONTEXT_CONTROL)
        {
            regs.ebp = context->Ebp;
            regs.esp = context->Esp;
            regs.eip = context->Eip;
            regs.cs  = context->SegCs;
            regs.ss  = context->SegSs;
            regs.eflags = context->EFlags;
        }
        if (flags & CONTEXT_SEGMENTS)
        {
            regs.ds = context->SegDs;
            regs.es = context->SegEs;
            regs.fs = context->SegFs;
            regs.gs = context->SegGs;
        }
        if (ptrace( PTRACE_SETREGS, tid, 0, &regs ) == -1) goto error;
    }
    if (flags & CONTEXT_DEBUG_REGISTERS)
    {
        if (ptrace( PTRACE_POKEUSER, tid, DR_OFFSET(0), context->Dr0 ) == -1) goto error;
        if (ptrace( PTRACE_POKEUSER, tid, DR_OFFSET(1), context->Dr1 ) == -1) goto error;
        if (ptrace( PTRACE_POKEUSER, tid, DR_OFFSET(2), context->Dr2 ) == -1) goto error;
        if (ptrace( PTRACE_POKEUSER, tid, DR_OFFSET(3), context->Dr3 ) == -1) goto error;
        if (ptrace( PTRACE_POKEUSER, tid, DR_OFFSET(6), context->Dr6 ) == -1) goto error;
        if (ptrace( PTRACE_POKEUSER, tid, DR_OFFSET(7), context->Dr7 ) == -1) goto error;
    }
    if (flags & CONTEXT_FLOATING_POINT)
    {
        /* we can use context->FloatSave directly as it is using the */
        /* correct structure (the same as fsave/frstor) */
        if (ptrace( PTRACE_SETFPREGS, tid, 0, &context->FloatSave ) == -1) goto error;
    }
    return;
 error:
    file_set_error();
}

#elif defined(__sun) || defined(__sun__)

/* retrieve a thread context */
void get_thread_context( struct thread *thread, unsigned int flags, CONTEXT *context )
{
    int tid = thread->unix_tid_or_pid;
    if (flags & CONTEXT_FULL)
    {
        struct regs regs;
        if (ptrace( PTRACE_GETREGS, tid, 0, (int) &regs ) == -1) goto error;
        if (flags & CONTEXT_INTEGER)
        {
            context->Eax = regs.r_eax;
            context->Ebx = regs.r_ebx;
            context->Ecx = regs.r_ecx;
            context->Edx = regs.r_edx;
            context->Esi = regs.r_esi;
            context->Edi = regs.r_edi;
        }
        if (flags & CONTEXT_CONTROL)
        {
            context->Ebp    = regs.r_ebp;
            context->Esp    = regs.r_esp;
            context->Eip    = regs.r_eip;
            context->SegCs  = regs.r_cs & 0xffff;
            context->SegSs  = regs.r_ss & 0xffff;
            context->EFlags = regs.r_efl;
        }
        if (flags & CONTEXT_SEGMENTS)
        {
            context->SegDs = regs.r_ds & 0xffff;
            context->SegEs = regs.r_es & 0xffff;
            context->SegFs = regs.r_fs & 0xffff;
            context->SegGs = regs.r_gs & 0xffff;
        }
    }
    if (flags & CONTEXT_DEBUG_REGISTERS)
    {
        /* FIXME: How is this done on Solaris? */
    }
    if (flags & CONTEXT_FLOATING_POINT)
    {
        /* we can use context->FloatSave directly as it is using the */
        /* correct structure (the same as fsave/frstor) */
        if (ptrace( PTRACE_GETFPREGS, tid, 0, (int) &context->FloatSave ) == -1) goto error;
        context->FloatSave.Cr0NpxState = 0;  /* FIXME */
    }
    return;
 error:
    file_set_error();
}


/* set a thread context */
void set_thread_context( struct thread *thread, unsigned int flags, const CONTEXT *context )
{
    int tid = thread->unix_tid_or_pid;
    if (flags & CONTEXT_FULL)
    {
        struct regs regs;
        if (((flags | CONTEXT_i386) & CONTEXT_FULL) != CONTEXT_FULL)
        {
            /* need to preserve some registers */
            if (ptrace( PTRACE_GETREGS, tid, 0, (int) &regs ) == -1) goto error;
        }
        if (flags & CONTEXT_INTEGER)
        {
            regs.r_eax = context->Eax;
            regs.r_ebx = context->Ebx;
            regs.r_ecx = context->Ecx;
            regs.r_edx = context->Edx;
            regs.r_esi = context->Esi;
            regs.r_edi = context->Edi;
        }
        if (flags & CONTEXT_CONTROL)
        {
            regs.r_ebp = context->Ebp;
            regs.r_esp = context->Esp;
            regs.r_eip = context->Eip;
            regs.r_cs = context->SegCs;
            regs.r_ss = context->SegSs;
            regs.r_efl = context->EFlags;
        }
        if (flags & CONTEXT_SEGMENTS)
        {
            regs.r_ds = context->SegDs;
            regs.r_es = context->SegEs;
            regs.r_fs = context->SegFs;
            regs.r_gs = context->SegGs;
        }
        if (ptrace( PTRACE_SETREGS, tid, 0, (int) &regs ) == -1) goto error;
    }
    if (flags & CONTEXT_DEBUG_REGISTERS)
    {
        /* FIXME: How is this done on Solaris? */
    }
    if (flags & CONTEXT_FLOATING_POINT)
    {
        /* we can use context->FloatSave directly as it is using the */
        /* correct structure (the same as fsave/frstor) */
        if (ptrace( PTRACE_SETFPREGS, tid, 0, (int) &context->FloatSave ) == -1) goto error;
    }
    return;
 error:
    file_set_error();
}

#elif defined(__FreeBSD__) || defined(__OpenBSD__)
#include <machine/reg.h>

/* retrieve a thread context */
void get_thread_context( struct thread *thread, unsigned int flags, CONTEXT *context )
{
    int tid = thread->unix_tid_or_pid;
    if (flags & CONTEXT_FULL)
    {
        struct reg regs;
        if (ptrace( PTRACE_GETREGS, tid, 0, (int) &regs ) == -1) goto error;
        if (flags & CONTEXT_INTEGER)
        {
            context->Eax = regs.r_eax;
            context->Ebx = regs.r_ebx;
            context->Ecx = regs.r_ecx;
            context->Edx = regs.r_edx;
            context->Esi = regs.r_esi;
            context->Edi = regs.r_edi;
        }
        if (flags & CONTEXT_CONTROL)
        {
            context->Ebp    = regs.r_ebp;
            context->Esp    = regs.r_esp;
            context->Eip    = regs.r_eip;
            context->SegCs  = regs.r_cs & 0xffff;
            context->SegSs  = regs.r_ss & 0xffff;
            context->EFlags = regs.r_eflags;
        }
        if (flags & CONTEXT_SEGMENTS)
        {
            context->SegDs = regs.r_ds & 0xffff;
            context->SegEs = regs.r_es & 0xffff;
            context->SegFs = regs.r_fs & 0xffff;
            context->SegGs = regs.r_gs & 0xffff;
        }
    }
    if (flags & CONTEXT_DEBUG_REGISTERS)
    {
        /* FIXME: How is this done on FreeBSD? */
    }
    if (flags & CONTEXT_FLOATING_POINT)
    {
        /* we can use context->FloatSave directly as it is using the */
        /* correct structure (the same as fsave/frstor) */
        if (ptrace( PTRACE_GETFPREGS, tid, 0, (int) &context->FloatSave ) == -1) goto error;
        context->FloatSave.Cr0NpxState = 0;  /* FIXME */
    }
    return;
 error:
    file_set_error();
}


/* set a thread context */
void set_thread_context( struct thread *thread, unsigned int flags, const CONTEXT *context )
{
    int tid = thread->unix_tid_or_pid;
    if (flags & CONTEXT_FULL)
    {
        struct reg regs;
        if (((flags | CONTEXT_i386) & CONTEXT_FULL) != CONTEXT_FULL)
        {
            /* need to preserve some registers */
            if (ptrace( PTRACE_GETREGS, tid, 0, (int) &regs ) == -1) goto error;
        }
        if (flags & CONTEXT_INTEGER)
        {
            regs.r_eax = context->Eax;
            regs.r_ebx = context->Ebx;
            regs.r_ecx = context->Ecx;
            regs.r_edx = context->Edx;
            regs.r_esi = context->Esi;
            regs.r_edi = context->Edi;
        }
        if (flags & CONTEXT_CONTROL)
        {
            regs.r_ebp = context->Ebp;
            regs.r_esp = context->Esp;
            regs.r_eip = context->Eip;
            regs.r_cs = context->SegCs;
            regs.r_ss = context->SegSs;
            regs.r_eflags = context->EFlags;
        }
        if (flags & CONTEXT_SEGMENTS)
        {
            regs.r_ds = context->SegDs;
            regs.r_es = context->SegEs;
            regs.r_fs = context->SegFs;
            regs.r_gs = context->SegGs;
        }
        if (ptrace( PTRACE_SETREGS, tid, 0, (int) &regs ) == -1) goto error;
    }
    if (flags & CONTEXT_DEBUG_REGISTERS)
    {
        /* FIXME: How is this done on FreeBSD? */
    }
    if (flags & CONTEXT_FLOATING_POINT)
    {
        /* we can use context->FloatSave directly as it is using the */
        /* correct structure (the same as fsave/frstor) */
        if (ptrace( PTRACE_SETFPREGS, tid, 0, (int) &context->FloatSave ) == -1) goto error;
    }
    return;
 error:
    file_set_error();
}


#elif defined(__APPLE__)
#ifdef MACH_X86_STATE_NEEDS_PREFIX
#define DREG(a) __##a
#else
#define DREG(a) a
#endif

/* retrieve a thread context */
static void get_thread_context (struct thread *thread, unsigned int flags,
                                CONTEXT *context)
{
   if (flags & CONTEXT_FULL)
   {
      x86_thread_state32_t regs;
      mach_msg_type_number_t sc = x86_THREAD_STATE32_COUNT;

      if (thread_get_state (thread->thread_port, x86_THREAD_STATE32,
                            (thread_state_t)&regs, &sc))
         goto error;

      if (flags & CONTEXT_INTEGER)
      {
         context->Eax = regs.DREG(eax);
         context->Ebx = regs.DREG(ebx);
         context->Ecx = regs.DREG(ecx);
         context->Edx = regs.DREG(edx);
         context->Esi = regs.DREG(esi);
         context->Edi = regs.DREG(edi);
      }

      if (flags & CONTEXT_CONTROL)
      {
         context->Ebp    = regs.DREG(ebp);
         context->Esp    = regs.DREG(esp);
         context->Eip    = regs.DREG(eip);
         context->SegCs  = regs.DREG(cs);
         context->SegSs  = regs.DREG(ss);
         context->EFlags = regs.DREG(eflags);
      }

      if (flags & CONTEXT_SEGMENTS)
      {
         context->SegDs = regs.DREG(ds);
         context->SegEs = regs.DREG(es);
         context->SegFs = regs.DREG(fs);
         context->SegGs = regs.DREG(gs);
      }
   }

   if (flags & CONTEXT_DEBUG_REGISTERS)
   {
      x86_debug_state32_t regs;
      mach_msg_type_number_t sc = x86_DEBUG_STATE32_COUNT;

      if (thread_get_state (thread->thread_port, x86_DEBUG_STATE32,
                            (thread_state_t)&regs, &sc))
         goto error;

      context->Dr0 = regs.DREG(dr0);
      context->Dr1 = regs.DREG(dr1);
      context->Dr2 = regs.DREG(dr2);
      context->Dr3 = regs.DREG(dr3);
      context->Dr6 = regs.DREG(dr6);
      context->Dr7 = regs.DREG(dr7);
   }

   if (flags & CONTEXT_FLOATING_POINT)
   {
      x86_float_state32_t regs;
      mach_msg_type_number_t sc = x86_FLOAT_STATE32_COUNT;
      unsigned short sVal;

      if (thread_get_state (thread->thread_port, x86_FLOAT_STATE32,
                            (thread_state_t)&regs, &sc))
         goto error;

      memcpy (&sVal, &regs.DREG(fpu_fcw), 2);
      context->FloatSave.ControlWord = sVal;

      memcpy (&sVal, &regs.DREG(fpu_fsw), 2);
      context->FloatSave.StatusWord = sVal;

      context->FloatSave.TagWord = regs.DREG(fpu_ftw);
      context->FloatSave.ErrorOffset = regs.DREG(fpu_ip);
      context->FloatSave.ErrorSelector = regs.DREG(fpu_cs);
      context->FloatSave.DataOffset = regs.DREG(fpu_dp);
      context->FloatSave.DataSelector = regs.DREG(fpu_ds);

      memcpy (context->FloatSave.RegisterArea, regs.DREG(fpu_stmm0).DREG(mmst_reg), 10);
      memcpy (context->FloatSave.RegisterArea + 10,
              regs.DREG(fpu_stmm1).DREG(mmst_reg), 10);
      memcpy (context->FloatSave.RegisterArea + 20,
              regs.DREG(fpu_stmm2).DREG(mmst_reg), 10);
      memcpy (context->FloatSave.RegisterArea + 30,
              regs.DREG(fpu_stmm3).DREG(mmst_reg), 10);
      memcpy (context->FloatSave.RegisterArea + 40,
              regs.DREG(fpu_stmm4).DREG(mmst_reg), 10);
      memcpy (context->FloatSave.RegisterArea + 50,
              regs.DREG(fpu_stmm5).DREG(mmst_reg), 10);
      memcpy (context->FloatSave.RegisterArea + 60,
              regs.DREG(fpu_stmm6).DREG(mmst_reg), 10);
      memcpy (context->FloatSave.RegisterArea + 70,
              regs.DREG(fpu_stmm7).DREG(mmst_reg), 10);

      context->FloatSave.Cr0NpxState = 0; /* FIXME */
   }

   return;
 error:
   set_error (STATUS_ACCESS_VIOLATION);
}


/* set a thread context */
static void set_thread_context (struct thread *thread, unsigned int flags,
                                const CONTEXT *context)
{
   if (flags & CONTEXT_FULL)
   {
      x86_thread_state32_t regs;
      mach_msg_type_number_t sc = x86_THREAD_STATE32_COUNT;

      if (thread_get_state (thread->thread_port, x86_THREAD_STATE32,
                            (thread_state_t)&regs, &sc))
         goto error;

      if (flags & CONTEXT_INTEGER)
      {
         regs.DREG(eax) = context->Eax;
         regs.DREG(ebx) = context->Ebx;
         regs.DREG(ecx) = context->Ecx;
         regs.DREG(edx) = context->Edx;
         regs.DREG(esi) = context->Esi;
         regs.DREG(edi) = context->Edi;
      }

      if (flags & CONTEXT_CONTROL)
      {
         regs.DREG(ebp) = context->Ebp;
         regs.DREG(esp) = context->Esp;
         regs.DREG(eip) = context->Eip;
         regs.DREG(cs) = context->SegCs;
         regs.DREG(ss) = context->SegSs;
         regs.DREG(eflags) = context->EFlags;
      }

      if (flags & CONTEXT_SEGMENTS)
      {
         regs.DREG(ds) = context->SegDs;
         regs.DREG(es) = context->SegEs;
         regs.DREG(fs) = context->SegFs;
         regs.DREG(gs) = context->SegGs;
      }

      if (thread_set_state (thread->thread_port, x86_THREAD_STATE32,
                            (thread_state_t)&regs, sc))
         goto error;
   }

   if (flags & CONTEXT_DEBUG_REGISTERS)
   {
      x86_debug_state32_t regs;
      mach_msg_type_number_t sc = x86_DEBUG_STATE32_COUNT;

      if (thread_get_state (thread->thread_port, x86_DEBUG_STATE32,
                            (thread_state_t)&regs, &sc))
         goto error;

      regs.DREG(dr0) = context->Dr0;
      regs.DREG(dr1) = context->Dr1;
      regs.DREG(dr2) = context->Dr2;
      regs.DREG(dr3) = context->Dr3;
      regs.DREG(dr6) = context->Dr6;
      regs.DREG(dr7) = context->Dr7;

      if (thread_set_state (thread->thread_port, x86_DEBUG_STATE32,
                            (thread_state_t)&regs, sc))
         goto error;
   }

   if (flags & CONTEXT_FLOATING_POINT)
   {
      x86_float_state32_t regs;
      mach_msg_type_number_t sc = x86_FLOAT_STATE32_COUNT;
      unsigned short sVal;

      if (thread_get_state (thread->thread_port, x86_FLOAT_STATE32,
                            (thread_state_t)&regs, &sc))
         goto error;

      sVal = context->FloatSave.ControlWord;
      memcpy (&regs.DREG(fpu_fcw), &sVal, 2);

      sVal = context->FloatSave.StatusWord;
      memcpy (&regs.DREG(fpu_fsw), &sVal, 2);

      regs.DREG(fpu_ftw) = context->FloatSave.TagWord;
      regs.DREG(fpu_ip) = context->FloatSave.ErrorOffset;
      regs.DREG(fpu_cs) = context->FloatSave.ErrorSelector;
      regs.DREG(fpu_dp) = context->FloatSave.DataOffset;
      regs.DREG(fpu_ds) = context->FloatSave.DataSelector;

      memcpy (regs.DREG(fpu_stmm0).DREG(mmst_reg), context->FloatSave.RegisterArea, 10);
      memcpy (regs.DREG(fpu_stmm1).DREG(mmst_reg),
              context->FloatSave.RegisterArea + 10, 10);
      memcpy (regs.DREG(fpu_stmm2).DREG(mmst_reg),
              context->FloatSave.RegisterArea + 20, 10);
      memcpy (regs.DREG(fpu_stmm3).DREG(mmst_reg),
              context->FloatSave.RegisterArea + 30, 10);
      memcpy (regs.DREG(fpu_stmm4).DREG(mmst_reg),
              context->FloatSave.RegisterArea + 40, 10);
      memcpy (regs.DREG(fpu_stmm5).DREG(mmst_reg),
              context->FloatSave.RegisterArea + 50, 10);
      memcpy (regs.DREG(fpu_stmm6).DREG(mmst_reg),
              context->FloatSave.RegisterArea + 60, 10);
      memcpy (regs.DREG(fpu_stmm7).DREG(mmst_reg),
              context->FloatSave.RegisterArea + 70, 10);

      if (thread_set_state (thread->thread_port, x86_FLOAT_STATE32,
                            (thread_state_t)&regs, sc))
         goto error;
   }

   return;
 error:
   set_error (STATUS_ACCESS_VIOLATION);
}

#else  /* __APPLE__ */
#error You must implement get/set_thread_context for your platform
#endif  /* linux || __sun__ || __FreeBSD__ || __APPLE__ */

/* copy a context structure according to the flags */
static void copy_context( CONTEXT *to, const CONTEXT *from, int flags )
{
    if (flags & CONTEXT_CONTROL)
    {
        to->Ebp    = from->Ebp;
        to->Eip    = from->Eip;
        to->Esp    = from->Esp;
        to->SegCs  = from->SegCs;
        to->SegSs  = from->SegSs;
        to->EFlags = from->EFlags;
    }
    if (flags & CONTEXT_INTEGER)
    {
        to->Eax = from->Eax;
        to->Ebx = from->Ebx;
        to->Ecx = from->Ecx;
        to->Edx = from->Edx;
        to->Esi = from->Esi;
        to->Edi = from->Edi;
    }
    if (flags & CONTEXT_SEGMENTS)
    {
        to->SegDs = from->SegDs;
        to->SegEs = from->SegEs;
        to->SegFs = from->SegFs;
        to->SegGs = from->SegGs;
    }
    if (flags & CONTEXT_FLOATING_POINT)
    {
        to->FloatSave = from->FloatSave;
    }
    /* we don't bother copying the debug registers, since they */
    /* always need to be accessed by ptrace anyway */
}

/* retrieve the current instruction pointer of a thread */
void *get_thread_ip( struct thread *thread )
{
    CONTEXT context;
    context.Eip = 0;
    if (suspend_for_ptrace( thread ))
    {
        get_thread_context( thread, CONTEXT_CONTROL, &context );
        resume_thread( thread );
    }
    return (void *)context.Eip;
}

/* determine if we should continue the thread in single-step mode */
int get_thread_single_step( struct thread *thread )
{
    CONTEXT context;
    if (thread->context) return 0;  /* don't single-step inside exception event */
    get_thread_context( thread, CONTEXT_CONTROL, &context );
    return (context.EFlags & 0x100) != 0;
}

/* retrieve the current context of a thread */
DECL_HANDLER(get_thread_context)
{
    struct thread *thread;
    void *data;
    int flags = req->flags & ~CONTEXT_i386;  /* get rid of CPU id */

    if (get_reply_max_size() < sizeof(CONTEXT))
    {
        set_error( STATUS_INVALID_PARAMETER );
        return;
    }
    if (!(thread = get_thread_from_handle( req->handle, THREAD_GET_CONTEXT ))) return;

    if ((data = set_reply_data_size( sizeof(CONTEXT) )))
    {
        /* copy incoming context into reply */
        memset( data, 0, sizeof(CONTEXT) );
        memcpy( data, get_req_data(), min( get_req_data_size(), sizeof(CONTEXT) ));

        if (thread->context)  /* thread is inside an exception event */
        {
            copy_context( data, thread->context, flags );
            flags &= CONTEXT_DEBUG_REGISTERS;
        }
        if (flags && suspend_for_ptrace( thread ))
        {
            get_thread_context( thread, flags, data );
            resume_thread( thread );
        }
    }
    release_object( thread );
}


/* set the current context of a thread */
DECL_HANDLER(set_thread_context)
{
    struct thread *thread;
    int flags = req->flags & ~CONTEXT_i386;  /* get rid of CPU id */

    if (get_req_data_size() < sizeof(CONTEXT))
    {
        set_error( STATUS_INVALID_PARAMETER );
        return;
    }
    if ((thread = get_thread_from_handle( req->handle, THREAD_SET_CONTEXT )))
    {
        if (thread->context)  /* thread is inside an exception event */
        {
            copy_context( thread->context, get_req_data(), flags );
            flags &= CONTEXT_DEBUG_REGISTERS;
        }
        if (flags && suspend_for_ptrace( thread ))
        {
            set_thread_context( thread, flags, get_req_data() );
            resume_thread( thread );
        }
        release_object( thread );
    }
}

#endif  /* __i386__ */
