/*
 * IO Completion Port support
 *
 * Copyright (C) 2007 TransGaming Technologies
 *
 * Main reference:
 *   http://www.microsoft.com/technet/sysinternals/information/IoCompletionPorts.mspx
 *
 * Status: IO completion ports are only partially implemented. In particular,
 * enough functionality is implemented to allow them to be used as task
 * queues.
 *
 * Known TODOs:
 * - keep num_active_threads current; we do proper accounting in here, but
 * also need to do so anywhere a thread can block (Sleep, WaitFor*, etc)
 * - implement associating a file handle (and its current thread) with a
 * completion port
 * - modify the following functions to be able to signal completion via
 * completion ports (and check around MSDN for any others):
    * ConnectNamedPipe
    * DeviceIoControl
    * LockFileEx
    * ReadDirectoryChangesW
    * ReadFile
    * TransactNamedPipe
    * WaitCommEvent
    * WriteFile
 */

#include "config.h"

#include <assert.h>
#include <stdio.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "winnt.h"
#include "handle.h"
#include "request.h"
#include "thread.h"


struct iocomp_task {
   struct iocomp_task *next;
   struct thread      *assigned_thread;
   void               *completion_key;
   void               *overlapped;
   unsigned int        num_bytes;
};


struct iocomp
{
   struct object   obj;                /* object header */
   unsigned int    max_active_threads; /* max number of concurrent threads */
   unsigned int    num_active_threads; /* num threads currently processing
                                          a task */
   int             abandon_waits;      /* used to prevent threads from
                                          getting signalled prematurely */
   struct iocomp_task *tasks;          /* ptr to head of FIFO task list */
   struct iocomp_task *last_task;      /* ptr to last task in FIFO task list */
   struct iocomp_task *assigned_tasks; /* ptr to head of assigned but not yet
                                          retrieved task list */
};


static void iocomp_dump (struct object *obj, int verbose);
static int iocomp_signaled (struct object *obj,
                            struct wait_queue_entry *entry);
static int iocomp_satisfied (struct object *obj, struct thread *thread);
static void iocomp_destroy (struct object *obj);

static const struct object_ops iocomp_ops =
{
   sizeof (struct iocomp),     /* size */
   iocomp_dump,                /* dump */
   add_queue,                  /* add_queue */
   remove_queue,               /* remove_queue */
   iocomp_signaled,            /* signaled */
   no_signal,                  /* signal */
   iocomp_satisfied,           /* satisfied */
   NULL,                       /* get_poll_events */
   NULL,                       /* poll_event */
   no_get_fd,                  /* get_fd */
   no_get_ctx_fd,              /* get_ctx_fd */
   no_flush,                   /* flush */
   no_get_file_info,           /* get_file_info */
   NULL,                       /* queue_async */
   iocomp_destroy              /* destroy */
};


const struct object_ops* get_this_process_iocomp_ops (void) {
   return &iocomp_ops;
}


static struct iocomp *create_iocomp (unsigned int MaxNumThreads)
{
   struct iocomp *pIOComp;

   if ((pIOComp = alloc_object (iocomp_ops_tag, -1)))
   {
      pIOComp->max_active_threads = MaxNumThreads;
      pIOComp->num_active_threads = 0;
      pIOComp->tasks = NULL;
      pIOComp->last_task = NULL;
      pIOComp->assigned_tasks = NULL;
      pIOComp->abandon_waits = 1;
   }

   return pIOComp;
}


static void iocomp_dump (struct object *obj, int verbose)
{
   struct iocomp *iocomp = (struct iocomp *)obj;
   struct iocomp_task *task;
   unsigned int tasks = 0, assigned_tasks = 0;

   task = iocomp->tasks;
   while (task)
   {
      tasks++;
      task = task->next;
   }

   task = iocomp->assigned_tasks;
   while (task)
   {
      assigned_tasks++;
      task = task->next;
   }

   assert (obj->ops == iocomp_ops_tag);
   fprintf (stderr, "IO Completion max=%u active=%u queued_tasks=%u assigned=%u ",
            iocomp->max_active_threads, iocomp->num_active_threads,
            tasks, assigned_tasks);
   dump_object_name (&iocomp->obj);
   fputc ('\n', stderr);
}


/* Return 1 if we have a task to be processed */
static int iocomp_signaled (struct object *obj,
                            struct wait_queue_entry *entry)
{
   struct iocomp *iocomp = (struct iocomp *)obj;
   struct iocomp_task *task;

   assert (obj->ops == iocomp_ops_tag);

   task = iocomp->assigned_tasks;
   if (task == NULL)
      return 0;

   while (task)
   {
      if (task->assigned_thread == entry->thread)
         return 1;
      task = task->next;
   }

   return 0;
}


/* Return 1 if it is safe for a waiting thread to be woken up */
static int iocomp_satisfied (struct object *obj, struct thread *thread)
{
   struct iocomp *iocomp = (struct iocomp *)obj;

   assert (obj->ops == iocomp_ops_tag);
   return iocomp->abandon_waits;
}


static void iocomp_destroy (struct object *obj)
{
   struct iocomp *iocomp = (struct iocomp *)obj;
   struct iocomp_task *task;

   assert (obj->ops == iocomp_ops_tag);

   while ((task = iocomp->tasks))
   {
      iocomp->tasks = task->next;
      mem_free (task);
   }

   while ((task = iocomp->assigned_tasks))
   {
      iocomp->assigned_tasks = task->next;
      mem_free (task);
   }
}


/* Selects the task at the top of the FIFO queue and assigns it to
   the passed thread (which then needs to retrieve it) */
static void assign_task_to_thread (struct iocomp *iocomp,
                                   struct thread *thread)
{
   struct iocomp_task *task;

   assert (iocomp->tasks != NULL);

   /* Remove task from tasks queue */
   task = iocomp->tasks;
   iocomp->tasks = task->next;
   if (iocomp->last_task == task)
      iocomp->last_task = NULL;

   /* Add task to assigned list, where it'll wait for the assigned thread
      to wake up and retrieve it */
   task->next = iocomp->assigned_tasks;
   iocomp->assigned_tasks = task;

   /* Assign it */
   task->assigned_thread = thread;

   /* Active thread accounting */
   iocomp->num_active_threads++;
   thread->iocomp = (struct iocomp *)grab_object (iocomp);
}


/* Kick off a thread if needed and available, and if we don't have the
   max number of threads running */
static void check_and_activate_thread (struct iocomp *iocomp)
{
   if (iocomp->tasks && iocomp->obj.tail &&
       (iocomp->num_active_threads < iocomp->max_active_threads))
   {
      struct wait_queue_entry * waiting = iocomp->obj.tail;

      assign_task_to_thread (iocomp, waiting->thread);
      wake_thread (waiting->thread);
   }
}


/* Notification routine called on thread exiting or blocking */
void iocomp_notify_thread_inactive (struct iocomp *iocomp)
{
   iocomp->num_active_threads--;

   /* Kick off a thread if appropriate */
   check_and_activate_thread (iocomp);
}


/* Notification routine called when a thread unblocks. Note that this can
   result in (temporarily) > max_active_threads running, but that's
   as expected */
void iocomp_notify_thread_active (struct iocomp *iocomp)
{
   iocomp->num_active_threads++;
}


DECL_HANDLER (create_io_completion)
{
   struct iocomp *iocomp;

   reply->handle = 0;

   if ((iocomp = create_iocomp (req->num_threads)))
   {
      reply->handle = alloc_handle (current->process, iocomp,
                                    IO_COMPLETION_ALL_ACCESS, 0);
      release_object (iocomp);
   }
}


DECL_HANDLER (set_io_completion)
{
   struct iocomp *iocomp;

   if ((iocomp =
        (struct iocomp *)get_handle_obj (current->process, req->handle,
                                         IO_COMPLETION_MODIFY_STATE,
                                         iocomp_ops_tag)))
   {
      struct iocomp_task *task;

      /* Create a new task and add it to the queue */
      task = mem_alloc (sizeof (struct iocomp_task));
      if (!task)
      {
         release_object (iocomp);
         return;
      }

      task->next = NULL;
      task->assigned_thread = NULL;
      task->completion_key = req->completion_key;
      task->overlapped = req->overlapped;
      task->num_bytes = req->num_bytes;

      if (iocomp->last_task)
         iocomp->last_task->next = task;
      else
         iocomp->tasks = task;
      iocomp->last_task = task;

      /* Kick off a thread if appropriate */
      check_and_activate_thread (iocomp);
      release_object (iocomp);
   }
}


DECL_HANDLER (remove_io_completion)
{
   struct iocomp *iocomp;

   /* Active thread accounting */
   if (current->iocomp)
   {
      current->iocomp->num_active_threads--;
      release_object (current->iocomp);
   }

   if ((iocomp =
        (struct iocomp *)get_handle_obj (current->process, req->handle,
                                         IO_COMPLETION_MODIFY_STATE,
                                         iocomp_ops_tag)))
   {
      /* Check for outstanding tasks; if so, and if there aren't too many
         threads already running, grab the one from the top of
         the FIFO queue and return */
      if (iocomp->tasks &&
          (iocomp->num_active_threads < iocomp->max_active_threads))
      {
         struct iocomp_task *task;

         /* Remove top task from queue */
         task = iocomp->tasks;
         iocomp->tasks = task->next;
         if (iocomp->last_task == task)
            iocomp->last_task = NULL;

         /* Return the task */
         reply->completion_key = task->completion_key;
         reply->overlapped = task->overlapped;
         reply->num_bytes = task->num_bytes;
         mem_free (task);

         /* Active thread accounting */
         iocomp->num_active_threads++;
         current->iocomp = (struct iocomp *)grab_object (iocomp);
      }
      /* No task. Time to wait for one if the timeout isn't infinite and
         is > 0 */
      else if (!(req->select_flags & SELECT_TIMEOUT) ||
               req->sec || req->usec)
      {
         struct timeval tv;

         /* Convert time to absolute time */
         gettimeofday (&tv, NULL);
         tv.tv_usec += req->usec;
         if (tv.tv_usec >= 1000000)
         {
            tv.tv_sec++;
            tv.tv_usec -= 1000000;
         }
         tv.tv_sec += req->sec;

         /* Need to ensure that existing waits on the port aren't
            interrupted while we're adding a new one */
         iocomp->abandon_waits = 0;
         select_on (1, req->cookie, &req->handle, req->select_flags,
                    tv.tv_sec, tv.tv_usec, 0, NULL);
         iocomp->abandon_waits = 1;

         reply->completion_key = NULL;
         reply->overlapped = NULL;
         reply->num_bytes = 0;
      }
      /* No task, and user specified timeout of 0 (meaning to return
         immediately) */
      else
      {
         set_error (STATUS_TIMEOUT);
         reply->completion_key = NULL;
         reply->overlapped = NULL;
         reply->num_bytes = 0;
      }

      release_object (iocomp);
   }
}


DECL_HANDLER (retrieve_assigned_io_completion)
{
   struct iocomp *iocomp;

   if ((iocomp =
        (struct iocomp *)get_handle_obj (current->process, req->handle,
                                         IO_COMPLETION_MODIFY_STATE,
                                         iocomp_ops_tag)))
   {
      struct iocomp_task *task, *prev;

      /* Walk assigned task list to find ours */
      task = iocomp->assigned_tasks;
      prev = NULL;
      while (task)
      {
         if (task->assigned_thread != current)
         {
            prev = task;
            task = task->next;
            continue;
         }

         break;
      }

      /* WTF? */
      if (!task)
      {
         fprintf (stderr,
                  "server: couldn't find assigned IO completion task?!\n");
         set_error (STATUS_INVALID_PARAMETER);
         reply->completion_key = NULL;
         reply->overlapped = NULL;
         reply->num_bytes = 0;
      }
      else
      {
         /* Remove task from list */
         if (prev)
            prev->next = task->next;
         else
            iocomp->assigned_tasks = task->next;

         reply->completion_key = task->completion_key;
         reply->overlapped = task->overlapped;
         reply->num_bytes = task->num_bytes;
         mem_free (task);
      }

      release_object (iocomp);
   }
}
