/*
 * Unicode string manipulation functions
 *
 * Copyright 2000 Alexandre Julliard
 * Copyright 2005 TransGaming Technologies, Inc
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "wine/unicode.h"

int strcmpiW( const WCHAR *str1, const WCHAR *str2 )
{
    for (;;)
    {
        int ret = toupperW(*str1) - toupperW(*str2);
        if (ret || !*str1) return ret;
        str1++;
        str2++;
    }
}

int strncmpiW( const WCHAR *str1, const WCHAR *str2, int n )
{
    int ret = 0;
    for ( ; n > 0; n--, str1++, str2++)
        if ((ret = toupperW(*str1) - toupperW(*str2)) || !*str1) break;
    return ret;
}

WCHAR *strstrW( const WCHAR *str, const WCHAR *sub )
{
    while (*str)
    {
        const WCHAR *p1 = str, *p2 = sub;
        while (*p1 && *p2 && *p1 == *p2) { p1++; p2++; }
        if (!*p2) return (WCHAR *)str;
        str++;
    }
    return NULL;
}


/* crappy implementations for strtolW and strtoulW
 * these need to be redone properly to parse the UNICODE directly,
 * instead of converting to UTF8 first, as this will only support
 * ASCII numbers */
long int strtolW( const WCHAR *str, WCHAR **end, int base ) 
{
    long int ret;
    int srclen = strlenW(str);
    int dstlen;
    char *str_utf8 = NULL;
    char *end_utf8;

    /* query size of utf8 string */
    dstlen = utf8_wcstombs(str, srclen, NULL, 0) + 1;

    /* make utf8 string, and ensure its null terminated */
    str_utf8 = (char *)malloc(dstlen*sizeof(char));
    utf8_wcstombs(str, srclen, str_utf8, dstlen);
    str_utf8[dstlen-1] = '\0';

    /* convert, and figure out the end */
    ret = strtol(str_utf8, &end_utf8, base);
    if( end ) {
      int len;
      /* figure out how much of the WCHAR string we used */
      len = utf8_mbstowcs(0, str_utf8, end_utf8 - str_utf8, NULL, 0);
      *end = (WCHAR *)str + len;
    }

    free(str_utf8);
    return ret;
}

unsigned long int strtoulW( const WCHAR *str, WCHAR **end, int base )
{
    unsigned long int ret;
    int srclen = strlenW(str);
    int dstlen;
    char *str_utf8 = NULL;
    char *end_utf8;

    /* query size of utf8 string */
    dstlen = utf8_wcstombs(str, srclen, NULL, 0) + 1;

    /* make utf8 string, and ensure its null terminated */
    str_utf8 = (char *)malloc(dstlen*sizeof(char));
    utf8_wcstombs(str, srclen, str_utf8, dstlen);
    str_utf8[dstlen-1] = '\0';

    /* convert, and figure out the end */
    ret = strtoul(str_utf8, &end_utf8, base);
    if( end ) {
      int len;
      /* figure out how much of the WCHAR string we used */
      len = utf8_mbstowcs(0, str_utf8, end_utf8 - str_utf8, NULL, 0);
      *end = (WCHAR *)str + len;
    }

    free(str_utf8);
    return ret;
}

int vsnprintfW( WCHAR *str, unsigned int len, const WCHAR *format, va_list valist )
{
  unsigned int written = 0;
  const WCHAR *iter = format;
  char bufa[256], fmtbufa[64], *fmta;

  while (*iter)
  {
    while (*iter && *iter != (WCHAR)L'%')
    {
     if (written++ >= len)
       return -1;
     *str++ = *iter++;
    }
    if (*iter == (WCHAR)L'%')
    {
      fmta = fmtbufa;
      *fmta++ = *iter++;
      while (*iter == (WCHAR)L'0' ||
             *iter == (WCHAR)L'+' ||
             *iter == (WCHAR)L'-' ||
             *iter == (WCHAR)L' ' ||
             *iter == (WCHAR)L'0' ||
             *iter == (WCHAR)L'*' ||
             *iter == (WCHAR)L'#')
      {
        if (*iter == (WCHAR)L'*')
        {
          char *buffiter = bufa;
          int fieldlen = va_arg(valist, int);
          sprintf(buffiter, "%d", fieldlen);
          while (*buffiter)
            *fmta++ = *buffiter++;
        }
        else
          *fmta++ = *iter;
        iter++;
      }

      while (isdigit(*iter))
        *fmta++ = *iter++;

      if (*iter == (WCHAR)L'.')
      {
        *fmta++ = *iter++;
        if (*iter == (WCHAR)L'*')
        {
          char *buffiter = bufa;
          int fieldlen = va_arg(valist, int);
          sprintf(buffiter, "%d", fieldlen);
          while (*buffiter)
            *fmta++ = *buffiter++;
        }
        else
          while (isdigit(*iter))
            *fmta++ = *iter++;
      }
      if (*iter == (WCHAR)L'h' ||
          *iter == (WCHAR)L'l')
      {
          *fmta++ = *iter++;
          *fmta++ = *iter++;
      }

      switch (*iter)
      {
      case (WCHAR)L's':
        {
          static const WCHAR none[] = { '(', 'n', 'u', 'l', 'l', ')', 0 };
          const WCHAR *wstr = va_arg(valist, const WCHAR *);
          const WCHAR *striter = wstr ? wstr : none;
          while (*striter)
          {
            if (written++ >= len)
              return -1;
            *str++ = *striter++;
          }
          iter++;
          break;
        }

      case (WCHAR)L'c':
        if (written++ >= len)
          return -1;
        *str++ = (WCHAR)va_arg(valist, int);
        iter++;
        break;

      default:
        {
          /* For non wc types, use system sprintf and append to wide char output */
          /* FIXME: for unrecognised types, should ignore % when printing */
          char *bufaiter = bufa;
          if (*iter == (WCHAR)L'p')
            sprintf(bufaiter, "%08lX", va_arg(valist, long));
          else
          {
            *fmta++ = *iter;
            *fmta = '\0';
            if (*iter == (WCHAR)L'f')
              sprintf(bufaiter, fmtbufa, va_arg(valist, double));
            else
              sprintf(bufaiter, fmtbufa, va_arg(valist, void *));
          }
          while (*bufaiter)
          {
            if (written++ >= len)
              return -1;
            *str++ = *bufaiter++;
          }
          iter++;
          break;
        }
      }
    }
  }
  if (written >= len)
    return -1;
  *str++ = (WCHAR)L'\0';
  return (int)written;
}

int vsprintfW( WCHAR* str, const WCHAR* format, va_list args )
{
  return vsnprintfW(str, INT_MAX, format, args);
}

int snprintfW( WCHAR *str, unsigned int len, const WCHAR *format, ... )
{
  int retval;
  va_list valist;
  va_start(valist, format);
  retval = vsnprintfW(str, len, format, valist);
  va_end(valist);
  return retval;
}

int sprintfW( WCHAR *str, const WCHAR *format, ... )
{
  int retval;
  va_list valist;
  va_start(valist, format);
  retval = vsnprintfW(str, INT_MAX, format, valist);
  va_end(valist);
  return retval;
}
